#!/usr/bin/perl
#
$USAGE = "Usage: $0 [-a, -f, -v] filename";
$argn = 0;

$verbose = 0;
$ascii   = 0;
$isTITLE = 0;
$isHTML  = 0;
$noFACE  = 0;
 
while ($ARGV[$argn] =~ /^\-/) {
  if ($ARGV[$argn] eq '-v') { $verbose = 1; }
  if ($ARGV[$argn] eq '-a') { $ascii   = 1; }
  if ($ARGV[$argn] eq '-f') { $noFACE  = 1; }
  $argn ++;
}

open (AAA, "<$ARGV[$argn]") or die $USAGE;
$text = "";
$lineindex[0] = 0;
$i = 0;
while (<AAA>) {
  # chop;
  $text .= $_;
  $i ++;
  $lineindex[$i] = length($text);
}
close AAA;

$len = $lineindex[$i];

$idir = rindex( $ARGV[$argn], '\/');
if ($idir >= 0) {
  $BASE = substr( $ARGV[$argn], 0, $idir);
} else {
  $BASE = "\.\/";
}

print "$ARGV[$argn] $i lines, $len chars, BASE $BASE\n";

if ($ascii) {
  do checkASCII($text);
}

$lb = index( $text, '<');
$rb = index( $text, '>');

$level = 0;
$error = 0;

while ( ($lb >= 0) && ($rb >= 0) ) {
  $comment = 0;
  $line = do linenumber( $lb );
  if ($lb > $rb) {
    $error = 1;
    # error
    print "$line: Error LB < RB : $lb $rb\n";
  } else {
    $str = substr($text, $lb, $rb-$lb);
    # print "$lb-$rb: SUBSTR $str \n";

    # skip comments <!-- ... -->
    if (substr($text, $lb+1, 3) eq '!--') {
      $comment = 1;
      $rbn = $rb;
      do {
        $str = substr($text, $rbn-2, 2);
        if ($str ne '--') {
          $rbn = index( $text, '>', $rbn+1); 
        }
      } while ($rbn > 0 && $str ne '--');
      if ($rbn < 0) {
        print "$line Error: runaway comment (starting at line $line)\n";
        $error = 1;
      }
      $lb = index( $text, '<', $rbn+1);
      $rb = index( $text, '>', $rbn+1); 
    } else {
      $lbn = index( $text, '<', $lb+1);
      if ($lbn < $rb) {
        $rbn = index( $text, '>', $rb+1); 
        if ($rbn >= 0) {
          $error = 1;
          # error: tag within tag
          print "$line: Error tag-in-tag at $lb \n";
        }  
      } else {
        $rbn = index( $text, '>', $rb+1);
      }
    }
    # $str = substr($text, $lbn, $rbn-$lbn);
    # print "$lb $rb --> $lbn $rbn $str\n";

  }

  last if $error;
  next if $comment;

  # get the tag
  if (substr($text, $lb+1, 1) eq "/") {   # closing
    $lb ++;
    $tag = substr($text, $lb+1, $rb-$lb-1);

    if ( ignoreTag($tag) == 0 ) {
      if ($verbose) { do printEndTag($line, $level, $tag); }
      if ($atag[$level] ne $tag) {
        print "$line Error: tag mismatch, line $line, tag $tag ";
        print "(expected $atag[$level] at line $aline[$level]) \n";
        $error = 1;
      }
      $level --;
    }
  } else {                                # opening
    $sp = index($text, ' ', $lb+1);
    $i = $sp;
    if ($sp < 0) { $sp = $len; }
    $nl = index( $text, '\n', $lb+1);
    $cr = index( $text, '\r', $lb+1);
    if ($nl > 0 && $nl < $sp) { $sp = $nl; }
    if ($cr > 0 && $cr < $sp) { $sp = $cr; }
    if ($rb > 0 && $rb < $sp) { $sp = $rb; }
    $tag = substr($text, $lb+1, $sp-$lb-1);
    $tag =~ s/\s//g;
    %attrh = ();
    if ($sp < $rb) {
      $attrs = substr($text, $sp+1, $rb-$sp-1);

      $attrs =~ s/^\s+//;
      $attrs =~ s/\s+$//;
      $attrs =~ s/\s*=\s*/=/g;
      $attrs =~ s/\s+/ /mg;

      @attrv = split(/ /, $attrs);
      for ($i=0; $i<@attrv; $i++) {
        if ($attrv[$i] =~ /=/) {
          ($key, $val) = split(/=/, $attrv[$i]);
          $val =~ s/\"//g;
          $val =~ s/\'//g;
        } else {
          $key = $attrv[$i];
          $val = "";
        } 
        $attrh{$key} = $val;
      }
    }
    if ( ignoreTag($tag) == 0 ) {
      $level ++;
      $atag[$level]=$tag;
      $aline[$level]=$line;
      if ($verbose) { do printStartTag($line, $level, $tag, \%attrh ); }
    } else {
      # printf "%5d: ", $line;
      # for ($i=1; $i<$level; $i++) { print ". "; }
      # print "  ($tag) \n";
    }
    do checkAttributes( $tag, $level, \%attrh );
  }
  
  $lb = $lbn;
  $rb = $rbn;

}

if ($level > 0) {
  print "$line Error: final-level $level\n";
} elsif (! $isHTML) {
  print "$line Error: no HTML tag\n";
} elsif (! $isTITLE) {
  print "$line Error: missing TITLE tag\n";
} elsif ($error) {
  print "*** Failed verify $ARGV[$argn]\n";
} else {
  print "*** Successful verify $ARGV[$argn]\n";
}

sub linenumber() {
  my $index = shift;
  my $i = 0;
  while ($lineindex[$i] < $index) { $i++; }
  return $i;
}

sub ignoreTag() {
  my $tg = shift;
  if (
        $tg ne "P"
     && $tg ne "BR" 
     && $tg ne "HR" 
     && $tg ne "META" 
     && $tg ne "FRAME" 
     && $tg ne "IMG"
     && $tg ne "INPUT"
     && $tg ne "AREA"
  ) { 
    return 0;
  } else {
    return 1;
  }
}

sub checkAttributes() {
  my $tg = shift;         # tag
  my $lv = shift;         # level
  my $rattrh = shift;     # attributes

  if ($lv == 1 && $tg ne 'HTML') {
    $error = 1;
    print "$line Error: missing start HTML tag (tag is $tg)\n";
    return 0;
  } else {
    $isHTML = 1;
  }

  if ($lv > 1 && $tg eq 'HTML') {
    $error = 1;
    print "$line Error: inner HTML tag at level $lv\n";
    return 0;
  }

  if ( # TAGS WITH NO ATTRIBUTE
       $tg eq 'H1'  || $tg eq 'H2'  || $tg eq 'H3' || $tg eq 'H4'
    || $tg eq 'B'   || $tg eq 'I'
    || $tg eq 'UL'  || $tg eq 'LI'  || $tg eq 'OL'
    || $tg eq 'SUB' || $tg eq 'SUP'
     ) {
    if (keys(%$rattrh) > 0) {
      print "$line Warning: $tg unexpected attributes\n";
      foreach $key (keys(%$rattrh)) {
        print "  $key => $$rattrh{$key} \n";
      }
    }
    return 0;
  } 
  if ($tg eq 'TITLE') {
    if ($lv == 3) {
      $isTITLE = 1;
    } else {
      $error = 1;
      print "$line Error: TITLE tag al level $lv (should be 3)\n";
    }
    return 0;
  } 
  if ($tg eq 'BODY') {
    # checkBODY($lv, $rattrh);
    return 0;
  } 
  if ($tg eq 'IMG') {
    return checkIMG($lv, $rattrh);
  } 
  if ($tg eq 'A') {
    return checkA($lv, $rattrh);
  } 
  if ($tg eq 'TABLE') {
    # checkTABLE($lv, $rattrh);
    return 0;
  } 
  if ($tg eq 'TR') {
    # checkTR($lv, $rattrh);
    return 0;
  } 
  if ($tg eq 'TH') {
    # checkTH($lv, $rattrh);
    return 0;
  } 
  if ($tg eq 'TD') {
    # checkTD($lv, $rattrh);
    return 0;
  } 
  if ($tg eq 'FONT') {
    return checkFONT($lv, $rattrh);
  } 
  if ($tg eq 'DIV') {
    # checkDIV($lv, $rattrh);
    return 0;
  } 
  if ($tg eq 'CENTER') {
    # checkCENTER($lv, $rattrh);
    return 0;
  }  
  return 0;
}

# check ANCHOR: maust be a link (HREF) or a NAME
#
sub checkA() {
  my $lv = shift;
  my $rattrh = shift;
  
  if ($$rattrh{'HREF'}) {
    $href = $$rattrh{'HREF'};
    if (index( $href, 'http') >= 0) {
      $target = $$rattrh{'TARGET'};
      $target =~ s/\"//g;
      if (! $target || $target ne "_top") {
        print STDERR "$line Error External link without TARGET=_top\n";
      }
    } elsif ( $href[0] != '\/' ) {
      $ref = $BASE . $href;
      if (! (-T $ref)) {
        $error = 1;
        print "$line Error: HREF $ref ($href) is not text file\n";
      } 
    }
  } elsif ($$rattrh{'NAME'}) {
    if (! $$rattrh{'VALUE'}) {
      $error = 1;
      print "$line Error: anchor NAME without VALUE\n";
    }
  } else {
    $error = 1;
    print "$line Error: anchor has neither HREF nor NAME\n";
  }
}

# image (IMG) must have SRC
#
sub checkIMG() {
  my $lv = shift;
  my $rattrh = shift;
  
  if (! $$rattrh{'SRC'}) {
    $error = 1;
    print "$line Error: IMG without SRC attribute\n";
    return 0;
  }
  # SHOULD CHECK ALSO WIDTH HEIGHT ... ???
  $src = $$rattrh{'SRC'};
  if (index( $src, 'http') >= 0 || $src =~ /^\// ) {
    print "$line Error: external image SRC $src \n";
  } else {
    $ref = $BASE . $src;
    if (! (-f $ref)) {
      $error = 1;
      print "$line Error: SRC $ref ($src) is not plain file\n";
    }
  } 
}

sub checkFONT() {
  my $lv = shift;
  my $rattrh = shift;
  if ($noFACE && $$rattrh{'FACE'}) {
    print STDERR "$line Error FONT with FACE = $$rattrh{'FACE'}\n";
  }
}

# the two printXXXTag are used in verbose mode
#
sub printStartTag() {
  my $ln = shift;
  my $lv = shift;
  my $tg = shift;
  my $rattrh = shift;

  printf "%5d: ", $ln;
  for ($i=1; $i<$lv; $i++) { print ". "; }
  print "< $tg ";
  if (%$rattrh) {
    print ": ";
    foreach $attr (keys(%$rattrh)) {
      print "$attr=>$$rattrh{$attr}  ";
    }
  }
  print "\n";
  return 0;
}

sub printEndTag() {
  my $ln = shift;
  my $lv = shift;
  my $tg = shift;

  printf "%5d: ", $ln;
  for ($i=1; $i<$lv; $i++) { print ". "; }
  print "> $tg \n";
  return 0;
}

# check that only 7-bit characters are used
#
sub checkASCII() {
  my $txt = shift;
  my $len = length( $txt );
  my $count= 0;
  for ($i=0; $i<$len; $i++) {
    if ( $txt[$i] & 0x80) { $count ++; }
  }
  if ($count > 0 ) {
    print "$ARGV[$argn] contains $count 8-bit chars\n";
  } else {
    print "$ARGV[$argn] is US-ASCII only\n";
  }
  return 0;
}

